package be.cylab.java.wowa.training;

import org.datavec.api.records.reader.RecordReader;
import org.datavec.api.records.reader.impl.csv.CSVRecordReader;
import org.datavec.api.split.FileSplit;
import org.datavec.api.util.ClassPathResource;
import org.deeplearning4j.datasets.datavec.RecordReaderDataSetIterator;
import org.deeplearning4j.eval.Evaluation;
import org.deeplearning4j.eval.ROC;
import org.deeplearning4j.nn.api.OptimizationAlgorithm;
import org.deeplearning4j.nn.conf.MultiLayerConfiguration;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.layers.DenseLayer;
import org.deeplearning4j.nn.conf.layers.OutputLayer;
import org.deeplearning4j.nn.multilayer.MultiLayerNetwork;
import org.deeplearning4j.nn.weights.WeightInit;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.DataSet;
import org.nd4j.linalg.dataset.SplitTestAndTrain;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.dataset.api.preprocessor.DataNormalization;
import org.nd4j.linalg.dataset.api.preprocessor.NormalizerStandardize;
import org.nd4j.linalg.factory.Nd4j;
import org.nd4j.linalg.lossfunctions.LossFunctions;

import java.io.IOException;
import java.util.List;

/**
 * Class to test neural network learning.
 */

public final class NeuralNetwork {

    /**
     * Class count.
     */
    public static final int CLASSES_COUNT = 2;
    /**
     * Features count.
     */
    public static final int FEATURES_COUNT = 5;


    private int neurons_number;
    private double learning_rate;
    private OptimizationAlgorithm algorithm;
    private Activation activation_function;


    /**
     * @param neurons_number
     * @param learning_rate
     * @param algorithm
     * @param activation_function
     */
    public NeuralNetwork(final int neurons_number,
                         final double learning_rate,
                         final OptimizationAlgorithm algorithm,
                         final Activation activation_function
    ) {

        this.learning_rate = learning_rate;
        this.algorithm = algorithm;
        this.activation_function = activation_function;
        this.neurons_number = neurons_number;
    }

    /**
     * Method to build a simple neural network, train and test it.
     *
     * @param neurons_number
     */
    MultiLayerNetwork learning(
            final int neurons_number,
            final DataSet learning,
            final DataSet testing) {

        MultiLayerConfiguration configuration
                = new NeuralNetConfiguration.Builder()
                .iterations(1000)
                .activation(this.activation_function)
                .optimizationAlgo(
                        this.algorithm)
                .weightInit(WeightInit.XAVIER)
                .learningRate(this.learning_rate)
                .regularization(true).l2(0.0001)
                .list()
                .layer(0, new DenseLayer.Builder().nIn(FEATURES_COUNT)
                        .nOut(neurons_number).build())
                .layer(1, new DenseLayer.Builder().nIn(neurons_number)
                        .nOut(neurons_number).build())
                .layer(2, new OutputLayer.Builder(LossFunctions
                        .LossFunction.NEGATIVELOGLIKELIHOOD)
                        .activation(Activation.SOFTMAX)
                        .nIn(neurons_number)
                        .nOut(CLASSES_COUNT).build())
                .backprop(true).pretrain(false)
                .build();

        MultiLayerNetwork model = new MultiLayerNetwork(configuration);
        model.init();
        model.fit(learning);

        INDArray output
                = model.output((testing.getFeatureMatrix()));

        Evaluation eval = new Evaluation(CLASSES_COUNT);
        eval.eval(testing.getLabels(), output);
        System.out.println(eval.stats());
        ROC roc = new ROC(CLASSES_COUNT);
        roc.eval(testing.getLabels(), output);
        System.out.println(roc.stats());
        return model;
    }

    /**
     * @param data
     * @param expected
     * @return
     */
    public MultiLayerNetwork run(
            final List<List<Double>> data,
            final List<Double> expected) {
        if (data.size() != expected.size()) {
            throw new IllegalArgumentException(
                    "Data and Expected must have the same size");
        }
        double[][] data_array = new double[data.size()][data.get(0).size()];
        double[][] expected_array = new double[expected.size()][2];
        for (int i = 0; i < data.size(); i++) {
            if (expected.get(i) == 1.0) {
                expected_array[i][0] = 1.0;
                expected_array[i][1] = 0.0;
            } else {
                expected_array[i][0] = 0.0;
                expected_array[i][1] = 1.0;
            }

            for (int j = 0; j < data.get(i).size(); j++) {
                data_array[i][j] = data.get(i).get(j);
            }
        }
        INDArray data_ind = Nd4j.create(data_array);
        INDArray expected_ind = Nd4j.create(expected_array);
        DataSet all_data = new DataSet(data_ind, expected_ind);
        all_data.shuffle();
        DataNormalization normalizer = new NormalizerStandardize();
        normalizer.fit(all_data);
        normalizer.transform(all_data);
        SplitTestAndTrain test_train = all_data.splitTestAndTrain(0.7);
        DataSet training_data = test_train.getTrain();
        DataSet testing_data  = test_train.getTest();
        return this.learning(neurons_number, training_data, testing_data);

    }

    /**
     * @param data_filename
     * @param expected_filename
     * @return
     */
    public MultiLayerNetwork run(
            final String data_filename,
            final String expected_filename) {
        List<List<Double>> data
                = Utils.convertJsonToDataForTrainer(data_filename);
        List<Double> expected
                = Utils.convertJsonToExpectedForTrainer(expected_filename);
        return this.run(data, expected);
    }

    /**
     * @param filename
     * @return
     */
    public MultiLayerNetwork runCSV(final String filename) {
        try (RecordReader record_reader = new CSVRecordReader(0, ',')) {
            record_reader.initialize(new FileSplit(
                    new ClassPathResource(filename).getFile()
            ));
            DataSetIterator iterator = new RecordReaderDataSetIterator(
                    record_reader, 12468, FEATURES_COUNT, CLASSES_COUNT);
            DataSet all_data = iterator.next();
            all_data.shuffle();
            DataNormalization normalizer = new NormalizerStandardize();
            normalizer.fit(all_data);
            normalizer.transform(all_data);

            SplitTestAndTrain test_and_train = all_data.splitTestAndTrain(0.80);
            DataSet training_data = test_and_train.getTrain();
            DataSet testing_data = test_and_train.getTest();
            return learning(neurons_number, training_data, testing_data);

        } catch (IOException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return null;
    }
}
